import logging
log = logging.getLogger('seantis.people')

from collections import namedtuple

import tablib

from plone import api

from zope.schema import getFields, getValidationErrors
from zope.schema.interfaces import (
    IFromUnicode, ValidationError, SchemaNotFullyImplemented
)

from seantis.people import _
from seantis.people.errors import ContentImportError
from seantis.plonetools import tools

supported_formats = ('csv', )


def import_people(container, portal_type, format, data):
    dataset = get_dataset(format, data)
    attribute_map = get_attribute_map(dataset.headers, portal_type)

    schema = tools.get_schema_from_portal_type(portal_type)

    for ix, record in enumerate(dataset.dict):
        try:
            values = get_attribute_values(record, attribute_map)
            validate_attribute_values(schema, values)

            api.content.create(
                container=container,
                type=portal_type,
                id='',
                **values
            )
        except ContentImportError, e:
            raise ContentImportError(
                e.message, rownumber=ix+1, colname=e.colname
            )
        except Exception, e:
            log.exception('error importing people')
            raise ContentImportError(e.message, rownumber=ix+1)

    return ix + 1


def get_dataset(format, data):
    assert format in supported_formats

    def get_dataset_from_csv():
        ds = tablib.Dataset()
        ds.csv = data

        return ds

    return locals()['get_dataset_from_{}'.format(format)]()


def get_attribute_map(headers, portal_type):

    if not headers:
        raise ContentImportError(_(u'No column headers were found'))

    schema = tools.get_schema_from_portal_type(portal_type)
    attribute_map = {}

    fields = getFields(schema)

    known_titles = dict((field.title, field) for field in fields.values())
    known_fields = dict((key, field) for key, field in fields.items())

    for header in headers:

        if header in known_fields:
            field = known_fields[header]
        elif header in known_titles:
            field = known_titles[header]
        else:
            continue

        if field in attribute_map.values():
            raise ContentImportError(
                _(
                    u'The ${name} column is specified more than once',
                    mapping=dict(name=header)
                )
            )

        attribute_map[header] = field

    return attribute_map


def get_attribute_values(record, attribute_map):
    values = {}

    for header, field in attribute_map.items():

        assert IFromUnicode.providedBy(field), """
            {} does not support fromUnicode
        """.format(field)

        try:
            values[field.__name__] = field.fromUnicode(record[header])
        except ValidationError, e:
            raise ContentImportError(e.doc(), colname=header)
        except ValueError, e:
            raise ContentImportError(e.message, colname=header)

    return values


def validate_attribute_values(schema, values):
    obj = namedtuple('ImportObject', values.keys())(**values)

    for error in getValidationErrors(schema, obj):
        if isinstance(error[1], SchemaNotFullyImplemented):
            raise ContentImportError(
                _(u'Required column is missing'), colname=error[0]
            )
        else:
            raise ContentImportError(
                error[1].message, colname=error[0]
            )



import re

from plone.supermodel import loadString, serializeSchema

from seantis.people import tests
from seantis.people.supermodel import (
    get_title_fields,
    set_title_fields,
    get_columns,
    set_columns,
    get_order,
    set_order,
    get_selectable_fields,
    set_selectable_fields,
    get_detail_fields,
    set_detail_fields
)


class TestSupermodel(tests.IntegrationTestCase):

    def deprettify(self, xml):
        return re.sub(' +', ' ', xml.replace('\n', '')).replace('> <', '><')

    title_xml = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:title>
                    <people:item>foo</people:item>
                    <people:item>bar</people:item>
                </people:title>
            </schema>
        </model>"""

    def test_load_title_schema(self):
        model = loadString(self.title_xml)
        self.assertEqual(
            get_title_fields(model.schema),
            ['foo', 'bar']
        )

    def test_write_title_schema(self):
        model = loadString(self.title_xml)
        set_title_fields(model.schema, ['stop', 'hammertime'])
        xml = serializeSchema(model.schema)

        self.assertIn('<people:item>stop</people:item>', xml)
        self.assertIn('<people:item>hammertime</people:item>', xml)
        self.assertTrue(
            xml.find('<people:item>stop</people:item>') <
            xml.find('<people:item>hammertime</people:item>')
        )

    column_xml = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column>
                        <people:item>
                            first

                        </people:item>
                        <people:item>   second </people:item>
                    </people:column>
                    <people:column>
                        <people:item> fourth </people:item>
                    </people:column>
                </people:columns>
            </schema>
        </model>"""

    def test_load_column_schema(self):
        model = loadString(self.column_xml)
        self.assertEqual(
            get_columns(model.schema),
            [['first', 'second'], ['fourth']]
        )

    def test_write_column_schema(self):
        model = loadString(self.column_xml)
        set_columns(
            model.schema, [['first'], ['third', 'fourth']]
        )
        xml = self.deprettify(serializeSchema(model.schema))

        self.assertIn('<people:column><people:item>first</people:item>', xml)
        self.assertIn('<people:column><people:item>third</people:item>', xml)
        self.assertIn('</people:item><people:item>fourth</people:item>', xml)

        self.assertTrue(
            xml.find('<people:item>first</people:item>') <
            xml.find('<people:item>third</people:item>') <
            xml.find('<people:item>fourth</people:item>')
        )

    order_xml = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:order>
                    <people:item>first</people:item>
                    <people:item>second</people:item>
                    <people:item>third</people:item>
                </people:order>
            </schema>
        </model>"""

    def test_load_order_schema(self):
        model = loadString(self.order_xml)
        self.assertEqual(
            list(get_order(model.schema)),
            ['first', 'second', 'third']
        )

    def test_write_order_schema(self):
        model = loadString(self.order_xml)
        set_order(model.schema, ['third', 'second', 'first'])
        xml = serializeSchema(model.schema)

        self.assertIn('<people:item>third</people:item>', xml)
        self.assertIn('<people:item>second</people:item>', xml)
        self.assertIn('<people:item>first</people:item>', xml)
        self.assertTrue(
            xml.find('<people:item>third</people:item>') <
            xml.find('<people:item>second</people:item>') <
            xml.find('<people:item>first</people:item>')
        )

    selectable_schema = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>works</people:item>
                    </people:column>
                    <people:column selectable="true">
                        <people:item>does</people:item>
                        <people:item>not</people:item>
                    </people:column>
                    <people:column>
                        <people:item>spare</people:item>
                    </people:column>
                </people:columns>
            </schema>
        </model>"""

    def test_load_selectable_schema(self):
        model = loadString(self.selectable_schema)

        # only colums with one field may be selectable
        self.assertEqual(
            sorted(get_selectable_fields(model.schema)),
            ['works']
        )

    def test_write_selectable_schema(self):
        model = loadString(self.selectable_schema)
        set_selectable_fields(model.schema, ['spare'])
        xml = self.deprettify(serializeSchema(model.schema))

        self.assertIn(
            '<people:column selectable="true"><people:item>spare', xml
        )

    details_schema = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:details>
                    <people:item>image</people:item>
                </people:details>
                <people:details position="right">
                    <people:item>lastname</people:item>
                    <people:item>firstname</people:item>
                </people:details>
            </schema>
        </model>"""

    def test_load_details_schema(self):
        model = loadString(self.details_schema)

        self.assertEqual(
            get_detail_fields(model.schema),
            {
                'left': ['image'],
                'right': ['lastname', 'firstname']
            }
        )

    def test_write_details_schema(self):
        model = loadString(self.details_schema)
        set_detail_fields(model.schema, {
            'top': ['image', 'lastname', 'firstname']
        })

        xml = self.deprettify(serializeSchema(model.schema))

        self.assertIn('<people:details position="top">', xml)
        self.assertIn('<people:item>lastname</people:item>', xml)
        self.assertIn('<people:item>firstname</people:item>', xml)

        self.assertTrue(
            xml.find('<people:item>lastname</people:item>') <
            xml.find('<people:item>firstname</people:item>')
        )

        self.assertEqual(xml.count('<people:details'), 1)



from plone import api

from Missing import MV

from seantis.people.supermodel import (
    get_columns,
    set_columns
)
from seantis.plonetools import tools

from seantis.people.supermodel.indexing import on_type_modified
from seantis.people.interfaces import IPerson
from seantis.people import tests


class TestColumns(tests.IntegrationTestCase):

    def test_add_person_column_first(self):
        new_type = self.new_temporary_type()

        set_columns(new_type.lookupSchema(), [['foo']])

        with self.user('admin'):
            obj = api.content.create(
                id='123',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        brain = tools.get_brain_by_object(obj)

        self.assertTrue(hasattr(brain, 'foo'))
        self.assertFalse(hasattr(brain, 'bar'))

        self.assertEqual(brain.foo, 'stop')

    def test_add_person_column_later(self):
        new_type = self.new_temporary_type()

        self.assertEqual(get_columns(new_type.lookupSchema()), [])

        with self.user('admin'):
            obj = api.content.create(
                id='123',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        # define the metadata after the obj has been created
        set_columns(new_type.lookupSchema(), [['foo']])

        # which leaves the attribute in a missing value state
        brain = tools.get_brain_by_object(obj)
        self.assertEqual(brain.foo, MV)

        # until reindexing happens
        obj.reindexObject()
        brain = tools.get_brain_by_object(obj)
        self.assertEqual(brain.foo, 'stop')

    def test_reindex_on_change(self):
        new_type = self.new_temporary_type(behaviors=[IPerson.__identifier__])

        set_columns(new_type.lookupSchema(), [['foo']])

        with self.user('admin'):
            obj = api.content.create(
                id='123',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        brain = tools.get_brain_by_object(obj)

        self.assertTrue(hasattr(brain, 'foo'))
        self.assertFalse(hasattr(brain, 'bar'))

        self.assertEqual(brain.foo, 'stop')

        set_columns(new_type.lookupSchema(), [['bar']])

        # usually the dexterity fti modified event does this
        on_type_modified(new_type)

        brain = tools.get_brain_by_object(obj)

        # The metadata is not deleted at this point because it is impossible
        # to tell if the metadata is used elsewhere. It's a rather big caveat..
        self.assertTrue(hasattr(brain, 'foo'))
        self.assertTrue(hasattr(brain, 'bar'))

        self.assertEqual(brain.bar, 'hammertime!')



from seantis.plonetools import setuphandlers

indexes = [
    ('first_letter', 'FieldIndex')
]


def add_catalog_indexes(context, logger=None):
    setuphandlers.add_catalog_indexes(
        'seantis.people', indexes, context, logger
    )


def import_indexes(context):
    setuphandlers.import_indexes(
        'seantis.people', indexes, context
    )



from seantis.plonetools import tools
from seantis.people.supermodel.security import has_read_access
from seantis.people.supermodel import (
    get_selectable_fields, get_columns, get_title_fields
)


def unrestricted_get_schema_columns(schema):
    return _get_schema_columns(schema, None, restricted=False)


def get_schema_columns(schema, context):
    return _get_schema_columns(schema, context, restricted=True)


def _get_schema_columns(schema, context, restricted):
    columns = []

    for fields in get_columns(schema):
        if restricted:
            fields = [
                f for f in fields if has_read_access(schema, f, context)
            ]

        if fields:
            columns.append(SchemaColumn(schema, fields))

    return columns


class SchemaColumn(object):

    def __init__(self, schema, fields):
        self.schema = schema
        self.fields = tools.order_fields_by_schema(fields, schema)

        self.titles = []
        for field in self.fields:
            if field in self.schema:
                self.titles.append(self.schema[field].title)
            else:
                self.titles.append(field)

        self.contains_title_field = bool(
            set(self.fields) & set(get_title_fields(self.schema))
        )
        self.selectable = (
            len(self.fields) == 1 and self.fields[0] in get_selectable_fields(
                self.schema
            )
        )




from plone.formwidget.contenttree import ObjPathSourceBinder
from plone.directives import form

from z3c.relationfield.schema import RelationChoice

from zope import schema
from zope.interface import Interface, invariant, Invalid

from seantis.people import _


class ISeantisPeopleSpecific(Interface):
    pass


class IPersonMarker(Interface):
    pass


class IPerson(Interface):

    def memberships(self):
        """ A list of membership belonging to this person. """

    def memberships_by_organizations(self, organizations=None):
        """ A dictionary of memberships belonging to this person. They key
        of the dictionary is the title of the organization. """


class INameFromPerson(Interface):
    pass


class IList(form.Schema):
    """ A list of IPerson objects. """

    title = schema.TextLine(
        title=_(u"Name of the list of people"),
        required=True
    )

    people_per_page = schema.Int(
        title=_(u"Number of people shown on each page"),
        required=True,
        default=50
    )


class IMembership(form.Schema):

    title = schema.TextLine(
        title=_(u"Role inside organization"),
        required=True
    )

    start = schema.Date(
        title=_(u"Start of membership"),
        required=False
    )

    end = schema.Date(
        title=_(u"End of membership"),
        required=False
    )

    person = RelationChoice(
        title=_(u"Person"),
        source=ObjPathSourceBinder(object_provides=IPerson.__identifier__),
        required=True,
    )

    @invariant
    def has_valid_daterange(Membership):
        if Membership.start is None:
            return

        if Membership.end is None:
            return

        if Membership.start > Membership.end:
            raise Invalid(_(u"The membership can't end before it starts"))



from five import grok

from seantis.plonetools import tools
from seantis.people.supermodel import security, get_detail_fields
from seantis.people.interfaces import IPersonMarker
from seantis.people.browser import BaseView, Renderer


class PersonView(BaseView):

    permission = 'zope2.View'
    grok.require(permission)
    grok.context(IPersonMarker)
    grok.name('view')

    template = grok.PageTemplateFile('templates/person.pt')

    def update(self):
        self.renderer = Renderer(self.schema)

    @property
    def schema(self):
        return tools.get_schema_from_portal_type(self.context.portal_type)

    def fields(self, position):
        has_access = lambda field: security.has_read_access(
            self.schema, field, self.context
        )
        fields = get_detail_fields(self.schema).get(position)

        if not fields:
            return []

        return [f for f in fields if has_access(f)]

    def get_field_title(self, field):
        if field in self.schema:
            return self.schema[field].title
        else:
            return field

    def visible_positions(self):
        positions = ('top', 'left', 'right', 'bottom')
        return [
            p for p in positions if any(
                getattr(self.context, f) for f in self.fields(p)
            )
        ]

    def split_screen(self):
        return set(('left', 'right')).issubset(set(self.visible_positions()))



from five import grok
from plone.directives.form import Form
from z3c.form.interfaces import ActionExecutionError
from zope.interface import Invalid
from seantis.people.interfaces import ISeantisPeopleSpecific


class BaseView(grok.View):

    grok.baseclass()
    grok.layer(ISeantisPeopleSpecific)


class BaseForm(Form):

    grok.baseclass()
    grok.layer(ISeantisPeopleSpecific)

    @property
    def parameters(self):
        data, errors = self.extractData()

        if errors:
            self.status = self.formErrorsMessage
            return None
        else:
            return data

    def raise_action_error(self, msg):
        raise ActionExecutionError(Invalid(msg))



from seantis.people import _
from seantis.plonetools import tools


class TranslatableException(Exception):

    def __repr__(self):
        return 'TranslatableException({})'.format(self.message)

    def __init__(self, message):
        self.message = message

    def translate(self, request):
        return tools.translator(request)(self.message)


class ContentImportError(TranslatableException):

    def __repr__(self):
        base = 'ContentImportError(message="{}", rownumber={}, colname="{}")'
        return base.format(self.message, self.rownumber, self.colname or '')

    def __init__(self, message, rownumber=None, colname=None):
        super(ContentImportError, self).__init__(message)
        self.rownumber = rownumber
        self.colname = colname

    def translate_position(self, request):
        particles = []

        if self.rownumber:
            particles.append(_(u'Row ${row}', mapping=dict(
                row=self.rownumber
            )))

        if self.colname:
            particles.append(_(u'Column ${column}', mapping=dict(
                column=self.colname
            )))

        translate = tools.translator(request)
        return u', '.join(map(translate, particles))

    def translate(self, request):
        message = super(ContentImportError, self).translate(request)
        position = self.translate_position(request)

        if not position:
            return message
        else:
            return '{} ({})'.format(message, position)



from plone import api

from seantis.people.interfaces import IPerson

from seantis.people.supermodel import set_detail_fields
from seantis.people import tests


class TestList(tests.IntegrationTestCase):

    def test_person_view_positions(self):
        with self.user('admin'):
            person_type = self.new_temporary_type(behaviors=[
                IPerson.__identifier__
            ])
            person = api.content.create(
                id='test',
                type=person_type.id,
                container=self.new_temporary_folder(),
                name='test',
                town='test',
                empty=''
            )
            schema = person_type.lookupSchema()
            view = person.unrestrictedTraverse('@@view')

        set_detail_fields(schema, {'top': ['name'], 'left': ['empty']})

        # empty positions are not visible and are therefore left out completely
        self.assertEqual(view.visible_positions(), ['top'])

        # split screen is available only if left and right are available
        self.assertFalse(view.split_screen())

        set_detail_fields(schema, {'left': ['name'], 'right': ['town']})

        self.assertEqual(view.visible_positions(), ['left', 'right'])
        self.assertTrue(view.split_screen())



from datetime import date
from plone import api
from zope.interface import Invalid

from seantis.people.interfaces import IMembership, IPerson
from seantis.people.content import Membership
from seantis.people import tests


class TestMembership(tests.IntegrationTestCase):

    def test_membership_creation(self):
        with self.user('admin'):
            ms = api.content.create(
                id='test',
                type='seantis.people.membership',
                container=self.new_temporary_folder()
            )

        self.assertIs(type(ms.aq_base), Membership)

    def test_membership_range_validation(self):
        with self.user('admin'):
            ms = api.content.create(
                id='test',
                type='seantis.people.membership',
                container=self.new_temporary_folder()
            )

        ms.start, ms.end = None, None
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), None
        IMembership.validateInvariants(ms)

        ms.start, ms.end = None, date(2012, 1, 1)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), date(2012, 1, 2)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 1), date(2012, 1, 1)
        IMembership.validateInvariants(ms)

        ms.start, ms.end = date(2012, 1, 2), date(2012, 1, 1)
        self.assertRaises(Invalid, IMembership.validateInvariants, ms)

    def test_membership_person_relation(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__]
        )

        with self.user('admin'):
            organization = self.new_temporary_folder('S.H.I.E.L.D')

            nick_fury = IPerson(api.content.create(
                title='Nick Fury',
                type=person_type.id,
                container=self.new_temporary_folder()
            ))

            api.content.create(
                title='Leader',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

        self.assertEqual(len(nick_fury.memberships()), 1)

        with self.user('admin'):
            api.content.create(
                title='Leutenant',
                type='seantis.people.membership',
                container=organization,
                person=nick_fury
            )

        self.assertEqual(len(nick_fury.memberships()), 2)

    def test_membership_by_organization(self):
        person_type = self.new_temporary_type(
            behaviors=[IPerson.__identifier__]
        )

        with self.user('admin'):
            uno = self.new_temporary_folder('UNO')
            nato = self.new_temporary_folder('NATO')

            john = IPerson(api.content.create(
                title='John Doe',
                type=person_type.id,
                container=self.new_temporary_folder()
            ))

            api.content.create(
                title='Janitor',
                type='seantis.people.membership',
                container=uno,
                person=john
            )

            api.content.create(
                title='Janitor',
                type='seantis.people.membership',
                container=nato,
                person=john
            )

        self.assertEqual(len(john.memberships()), 2)
        self.assertEqual(len(john.memberships_by_organization()), 2)
        self.assertEqual(len(john.memberships_by_organization(['UNO'])), 1)
        self.assertEqual(len(john.memberships_by_organization(['NATO'])), 1)
        self.assertEqual(len(john.memberships_by_organization(['ASDF'])), 0)



from seantis.people import _
from seantis.people import errors, tests


class TestErrors(tests.IntegrationTestCase):

    def test_error_translation(self):
        err = errors.TranslatableException(_(u'Test'))
        self.assertEqual(err.translate(self.request), u'Test')

        err = errors.ContentImportError(_(u'Test'))
        self.assertEqual(err.translate(self.request), u'Test')

        err.colname = 'test'
        self.assertEqual(err.translate(self.request), u'Test (Column test)')

        err.rownumber = 1
        self.assertEqual(
            err.translate(self.request), u'Test (Row 1, Column test)'
        )



# -*- coding: utf-8 -*-
from plone import api
from plone.indexer import indexer

from seantis.plonetools import tools

from seantis.people.interfaces import IPerson, IPersonMarker
from seantis.people.supermodel import (
    get_selectable_fields, get_order
)


def get_sortable_title_text(obj):
    schema = tools.get_schema_from_portal_type(obj.portal_type)
    order = list(get_order(schema))

    if order:
        title = u' '.join((getattr(obj, field, u'') for field in order))
        title = title.strip()

    if not order or not title:
        title = getattr(obj, 'title', u'')

    return title.strip()


@indexer(IPersonMarker)
def sortable_title(obj):
    return tools.unicode_collate_sortkey()(get_sortable_title_text(obj))


@indexer(IPersonMarker)
def first_letter(obj):
    title = get_sortable_title_text(obj)
    return title and title[:1].upper() or u''


def on_type_modified(fti, event=None):
    """ The IPerson types need to be reindexed if the type changes, because
    the supermodel could be different and it's hints may have an effect on
    the metadata/indexes.

    """

    update_related_indexes(fti)


def update_related_indexes(fti):

    if IPerson.__identifier__ not in fti.behaviors:
        return

    catalog = api.portal.get_tool('portal_catalog')
    new_indexes = update_selectable_field_indexes(fti)

    if new_indexes:
        for new_index in new_indexes:
            catalog.reindexIndex(new_index, REQUEST=None)

    for brain in catalog(portal_type=fti.id):
        brain.getObject().reindexObject(
            idxs=['sortable_title', 'first_letter']
        )


def get_selectable_prefix(portal_type):
    return portal_type + '_selectable_'


def get_selectable_field_ix(portal_type, field):
    return get_selectable_prefix(portal_type) + field


def get_selectable_field_indexes(fti):
    prefix = get_selectable_prefix(fti.id)
    zcatalog = api.portal.get_tool('portal_catalog')._catalog
    return [ix for ix in zcatalog.indexes if ix.startswith(prefix)]


def update_selectable_field_indexes(fti):
    catalog = api.portal.get_tool('portal_catalog')
    fields = get_selectable_fields(fti.lookupSchema())

    new_indexes = []

    # remove the indexes which are no longer used
    prefix = get_selectable_prefix(fti.id)
    for ix in get_selectable_field_indexes(fti):
        field = ix.replace(prefix, '')

        if field not in fields:
            catalog.delIndex(ix)

    # add the indexes which are not yet defined
    for field in fields:
        index_name = get_selectable_field_ix(fti.id, field)
        if index_name not in catalog.indexes():
            catalog.addIndex(index_name, 'FieldIndex', extra={
                'indexed_attrs': field
            })
            new_indexes.append(index_name)

    return new_indexes



from collections import namedtuple

from zope.interface import directlyProvides

from Products.ZCatalog.interfaces import ICatalogBrain

from plone.supermodel import loadString
from plone.namedfile.file import NamedImage

from seantis.people.browser.renderer import Renderer
from seantis.people import tests


class TestRenderer(tests.IntegrationTestCase):

    model = """<?xml version='1.0' encoding='utf8'?>
    <model xmlns="http://namespaces.plone.org/supermodel/schema">
        <schema>
            <field name="textline"
                   type="zope.schema.TextLine">
                <title>Textline</title>
            </field>
            <field name="email"
                   type="seantis.plonetools.schemafields.Email">
                <title>Email</title>
            </field>
            <field name="website"
                   type="seantis.plonetools.schemafields.Website">
                <title>Website</title>
            </field>
            <field name="blobimage"
                   type="plone.namedfile.field.NamedBlobImage">
                <title>Blobimage</title>
            </field>
            <field name="image"
                   type="plone.namedfile.field.NamedImage">
                <title>Image</title>
            </field>
        </schema>
    </model>"""

    @property
    def schema(self):
        return loadString(self.model).schema

    def render_value(self, field, value):
        renderer = Renderer(self.schema)
        context = namedtuple('MockContext', [field])(value)
        return renderer.render(context, field)

    def test_textline(self):
        self.assertEqual(self.render_value('textline', u'test'), u'test')

    def test_email(self):
        self.assertEqual(
            self.render_value('email', u'test@example.com'),
            u'<a href="mailto:test@example.com">test@example.com</a>'
        )

    def test_website(self):
        self.assertEqual(
            self.render_value('website', u'http://example.com'),
            (
                u'<a href="http://example.com" '
                u'target="_blank">http://example.com</a>'
            )
        )

    def test_image(self):

        renderer = Renderer(self.schema)

        class MockContext(object):
            def getURL(self):
                return u'http://nohost/mockbrain'

            def absolute_url(self):
                return u'http://nohost/mockobject'

        context = MockContext()
        context.image = None
        self.assertEqual(renderer.render(context, 'image'), '')

        # if the context does not provide ICatalogBrain, absolute_url is used
        context.image = NamedImage()
        self.assertEqual(
            renderer.render(context, 'image'),
            u'<img src="http://nohost/mockobject/@@images/image/thumb" />'
        )

        # if it does, getURL is used
        directlyProvides(context, ICatalogBrain)
        self.assertEqual(
            renderer.render(context, 'image'),
            u'<img src="http://nohost/mockbrain/@@images/image/thumb" />'
        )



""" Provides a fast renderer for schema fields. Since this renderer is used
on lists and tables and therefore called many times in a request the following
options were not chosen:

z3c.form display mode rendering - too complicated, hard to customize and slow
zpt templates - quite a bit slower because of features we don't need

The snippets rendered here so simple that string.Template can be used which
is by far the fastest way of doing templates in python.

"""

import string

from zope.schema import getFields
from plone.namedfile.field import NamedBlobImage, NamedImage
from Products.ZCatalog.interfaces import ICatalogBrain

from seantis.plonetools.schemafields import Email, Website


class EmailFieldRenderer(object):

    template = string.Template(u'<a href="mailto:${url}">${url}</a>')

    def __call__(self, context, field):
        return self.template.substitute(url=getattr(context, field))


class WebsiteFieldRenderer(object):

    template = string.Template(u'<a href="${url}" target="_blank">${url}</a>')

    def __call__(self, context, field):
        return self.template.substitute(url=getattr(context, field))


class ImageRenderer(object):

    template = string.Template(u'<img src="${url}" />')

    def __call__(self, context, field):
        img = getattr(context, field)
        if img:
            if ICatalogBrain.providedBy(context):
                baseurl = context.getURL()
            else:
                baseurl = context.absolute_url()

            url = '/'.join((baseurl, '@@images', field, 'thumb'))

            return self.template.substitute(url=url)
        else:
            return u''


renderers = {
    Email: EmailFieldRenderer(),
    Website: WebsiteFieldRenderer(),
    NamedBlobImage: ImageRenderer(),
    NamedImage: ImageRenderer()
}


class Renderer(object):

    def __init__(self, schema):
        self.schema = schema
        self.fields = getFields(schema)

    def render(self, context, field):
        fieldtype = type(self.fields.get(field, None))
        return renderers.get(fieldtype, getattr)(context, field)



from plone import api
from plone.browserlayer.utils import registered_layers

from seantis.people import tests
from seantis.people import interfaces


class TestSetup(tests.IntegrationTestCase):

    def test_profile_installed(self):
        setup = api.portal.get_tool('portal_setup')
        version = setup.getLastVersionForProfile('seantis.people:default')

        self.assertNotEqual(version, None)
        self.assertNotEqual(version, 'unknown')

    def test_request_layer_active(self):
        layers = registered_layers()
        self.assertIn(interfaces.ISeantisPeopleSpecific, layers)



import transaction
from copy import copy
from five import grok

from zope import schema
from z3c.form import field
from z3c.form.button import buttonAndHandler
from plone.directives import form
from plone.namedfile.field import NamedFile

from seantis.people import _
from seantis.people.errors import ContentImportError
from seantis.people.interfaces import IList
from seantis.people.browser import BaseForm
from seantis.people.content.import_content import (
    supported_formats,
    import_people
)

from zope.schema.vocabulary import SimpleVocabulary


class IImportFormSchema(form.Schema):

    portal_type = schema.Choice(title=_(u"Type"), values=[])
    import_file = NamedFile(title=_(u"File"))


class ImportForm(BaseForm):

    label = _(u'Import people')

    permission = 'cmf.ManagePortal'
    grok.require(permission)
    grok.context(IList)
    grok.name('import')

    ignoreContext = True

    template = grok.PageTemplateFile('templates/import_form.pt')

    @property
    def fields(self):
        fields = field.Fields(IImportFormSchema)

        for f in fields.values():
            f.field = copy(f.field)

        fields['portal_type'].field.vocabulary = SimpleVocabulary.fromItems(
            (fti.title, fti.id) for fti in self.context.available_types()
        )

        return fields

    def get_format_from_filename(self, filename):
        if not '.' in filename:
            return ''
        else:
            return filename.split('.')[-1]

    @buttonAndHandler(_(u'Import'), name='import')
    def run_import(self, action):

        if not self.parameters:
            return

        try:
            portal_type = self.parameters['portal_type']
            import_file = self.parameters['import_file']

            format = self.get_format_from_filename(import_file.filename)

            if not format in supported_formats:
                raise ContentImportError(_(
                    u'${format} is not supported, supported are ${formats}',
                    mapping={
                        'format': format,
                        'formats': ', '.join(supported_formats)
                    }
                ))

            count = import_people(
                self.context, portal_type, format, import_file.data
            )

            self.status = _(
                u'Successfully imported ${count} records', mapping=dict(
                    count=count
                )
            )

        except ContentImportError, e:
            transaction.abort()
            self.raise_action_error(e.translate(self.request))

    @buttonAndHandler(_(u'Cancel'), name='cancel')
    def cancel(self, action):
        self.request.response.redirect(self.context.absolute_url())



from plone.supermodel import loadString

from seantis.people import tests
from seantis.people.supermodel import security
from seantis.people.supermodel import (
    unrestricted_get_schema_columns,
    get_schema_columns,
    set_columns,
    set_title_fields
)


class TestSchemaColumns(tests.IntegrationTestCase):

    address_model = """<?xml version='1.0' encoding='utf8'?>
        <model  xmlns="http://namespaces.plone.org/supermodel/schema"
                xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <field  name="lastname" type="zope.schema.TextLine">
                    <title>Nachname</title>
                </field>
                <field  name="firstname" type="zope.schema.TextLine">
                    <title>Vorname</title>
                </field>
                <field  name="town" type="zope.schema.TextLine">
                    <title>Ort</title>
                </field>
                <field  name="country" type="zope.schema.TextLine">
                    <title>Land</title>
                </field>
            </schema>
        </model>"""

    def test_no_columns(self):

        model = loadString(self.address_model)
        columns = unrestricted_get_schema_columns(model.schema)

        self.assertEqual(columns, [])

    def test_one_column(self):

        model = loadString(self.address_model)

        set_columns(model.schema, [['lastname']])
        columns = unrestricted_get_schema_columns(model.schema)

        self.assertEqual(len(columns), 1)
        self.assertEqual(columns[0].schema, model.schema)
        self.assertEqual(columns[0].fields, ['lastname'])
        self.assertEqual(columns[0].titles, [u'Nachname'])
        self.assertEqual(columns[0].contains_title_field, False)

    def test_merged_columns(self):

        model = loadString(self.address_model)

        set_columns(model.schema, [['country', 'lastname']])
        columns = unrestricted_get_schema_columns(model.schema)

        self.assertEqual(len(columns), 1)

        # ordered by field order
        self.assertEqual(columns[0].fields, ['lastname', 'country'])
        self.assertEqual(columns[0].titles, [u'Nachname', u'Land'])
        self.assertEqual(columns[0].contains_title_field, False)

    def test_contains_title_field(self):

        model = loadString(self.address_model)

        set_columns(model.schema, [['country', 'lastname']])

        columns = unrestricted_get_schema_columns(model.schema)
        self.assertEqual(columns[0].contains_title_field, False)

        set_title_fields(model.schema, ['lastname'])
        columns = unrestricted_get_schema_columns(model.schema)
        self.assertEqual(columns[0].contains_title_field, True)

        set_title_fields(model.schema, ['firstname'])
        columns = unrestricted_get_schema_columns(model.schema)
        self.assertEqual(columns[0].contains_title_field, False)

        set_columns(model.schema, [['firstname']])
        columns = unrestricted_get_schema_columns(model.schema)
        self.assertEqual(columns[0].contains_title_field, True)

    def test_restricted_columns(self):

        context = self.new_temporary_folder()
        model = loadString(self.address_model)

        set_columns(model.schema, [
            ['firstname', 'lastname'],
            ['town'],
            ['country'],
        ])
        security.set_read_permissions(model.schema, {
            'firstname': 'cmf.ManagePortal',
            'lastname': 'zope2.View'
        })

        # anonymous
        columns = get_schema_columns(model.schema, context)
        self.assertEqual(len(columns), 3)
        self.assertEqual(columns[0].fields, ['lastname'])

        # admin
        with self.user('admin'):
            columns = get_schema_columns(model.schema, context)

        self.assertEqual(len(columns), 3)
        self.assertEqual(columns[0].fields, ['lastname', 'firstname'])



import logging
log = logging.getLogger('seantis.people')

from zipfile import ZipFile

import csv
import os.path
import inspect

from StringIO import StringIO
from urllib2 import urlopen

from five import grok
from plone import api

from seantis.people.browser import BaseView
from seantis.people.interfaces import IList
from seantis.plonetools import tools


class LoadTestRecordsView(BaseView):

    permission = 'cmf.ManagePortal'
    grok.require(permission)

    grok.context(IList)
    grok.name('load-test-records')

    typename = 'test-record'

    @property
    def count(self):
        return self.request.get('count', '500')

    @property
    def url(self):
        return 'http://www.briandunning.com/sample-data/{}.zip'.format(
            self.count
        )

    @property
    def csv(self):
        return '{}.csv'.format(self.count)

    def load_xml(self):
        path = os.path.dirname(
            os.path.abspath(
                inspect.getfile(inspect.currentframe())
            )
        )

        return open(os.path.join(path, 'load_test_records.xml'), 'r').read()

    def create_type(self):
        if tools.is_existing_portal_type(self.typename):
            return

        tools.add_new_dexterity_type(
            self.typename, model_source=self.load_xml(),
            title=u'Test Record',
            behaviors=[
                'seantis.people.interfaces.IPerson',
                'seantis.people.interfaces.INameFromPerson'
            ]
        )

    def import_sample_data(self):
        zipfile = ZipFile(StringIO(urlopen(self.url).read()))

        # get rid of windows newlines - there's probably a more efficient way
        content = '\n'.join(zipfile.open(self.csv).read().splitlines())
        reader = csv.reader(StringIO(content), delimiter=',')

        headers = None

        try:
            from collective.noindexing.patches import apply, unapply
            noindexing_available = True
            log.info("using collective.noindexing to speed the import up")
        except ImportError:
            noindexing_available = False
            log.info("install collective.noindexing for faster imports")

        if noindexing_available:
            apply()

        for rowix, row in enumerate(reader):
            if not headers:
                headers = [key.lower() for key in row]
            else:
                kwargs = dict(
                    (header, row[ix]) for ix, header in enumerate(headers)
                )

                api.content.create(
                    id='',
                    type=self.typename,
                    container=self.context,
                    **kwargs
                )

            if rowix % 100 == 0:
                log.info('imported 100 records')

        if noindexing_available:
            log.info('rebuilding the catalog')
            unapply()
            api.portal.get_tool('portal_catalog').clearFindAndRebuild()

        return rowix

    def render(self):
        if not tools.in_debug_mode():
            return "zope must be run in debug mode for this to work"

        RESPONSE = self.request.RESPONSE
        RESPONSE.setHeader('Content-Type', 'text/plain;charset=utf-8')

        log.info('creating type')
        self.create_type()

        log.info('importing sample data')
        rows = self.import_sample_data()

        log.info('imported {} records'.format(rows))
        self.request.response.redirect(self.context.absolute_url())



from plone import api

from seantis.people.supermodel import (
    set_order
)
from seantis.plonetools import tools

from seantis.people.interfaces import IPerson
from seantis.people import tests


class TestOrder(tests.IntegrationTestCase):

    def test_order(self):
        new_type = self.new_temporary_type(behaviors=[IPerson.__identifier__])
        set_order(new_type.lookupSchema(), ['bar', 'foo'])

        with self.user('admin'):
            obj = api.content.create(
                id='123',
                type=new_type.id,
                container=self.new_temporary_folder(),
                foo='stop',
                bar='hammertime!'
            )

        brain = tools.get_brain_by_object(obj)
        catalog = api.portal.get_tool('portal_catalog')

        self.assertEqual(
            catalog.getIndexDataForRID(brain.getRID())['sortable_title'],
            tools.unicode_collate_sortkey()('hammertime! stop')
        )



# -*- coding: utf-8 -*-

from plone import api

from Products.CMFPlone.interfaces.constrains import IConstrainTypes, ENABLED

from seantis.people.interfaces import IPerson
from seantis.people.content import List
from seantis.people import tests
from seantis.people.content.list import ListConstrainTypes, LetterFilter

from pyquery import PyQuery as pq


class TestList(tests.IntegrationTestCase):

    def test_list_creation(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

        self.assertIs(type(lst.aq_base), List)

    def test_list_people_by_behavior(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            new_person_type = lambda: self.new_temporary_type(
                behaviors=[IPerson.__identifier__]
            )

            person = new_person_type()
            new_person_type()  # intentionally unused

            # all types are possible before any are used
            self.assertEqual(len(lst.possible_types()), 2)
            self.assertEqual(len(lst.available_types()), 2)
            self.assertEqual(lst.used_type(), None)

            api.content.create(id='ned', type=person.id, container=lst)
            api.content.create(id='robb', type=person.id, container=lst)

            # only one type is used at this point
            self.assertEqual(len(lst.people()), 2)
            self.assertEqual(len(lst.possible_types()), 2)
            self.assertEqual(len(lst.available_types()), 1)
            self.assertEqual(lst.used_type(), person)

    def test_list_constrain(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )
            constrain = IConstrainTypes(lst)

            new_type = self.new_temporary_type(
                behaviors=[IPerson.__identifier__]
            )

            self.assertIs(type(constrain), ListConstrainTypes)
            self.assertEqual(constrain.getConstrainTypesMode(), ENABLED)
            self.assertEqual(constrain.allowedContentTypes(), [new_type])
            self.assertEqual(constrain.getLocallyAllowedTypes(), [new_type.id])
            self.assertEqual(
                constrain.getImmediatelyAddableTypes(), [new_type.id]
            )

    def test_add_unsupported_type(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            add_invalid_type = lambda: api.content.create(
                id='invalid', type=self.new_temporary_type(), container=lst
            )

        self.assertRaises(api.exc.InvalidParameterError, add_invalid_type)

    def test_list_letters(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__]
            )

            api.content.create(title='Jessie', type=person.id, container=lst)
            api.content.create(title='Walt', type=person.id, container=lst)

        self.assertEqual(sorted(lst.letters()), ['J', 'W'])

        filtered = lst.people(filter=LetterFilter('J', None))
        self.assertEqual(len(filtered), 1)
        self.assertEqual(filtered[0].id, 'jessie')

    def test_list_view_empty(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        self.assertIn('No people in the list.', view())

    def test_list_view_columns(self):

        model = """<?xml version='1.0' encoding='utf8'?>
        <model
            xmlns="http://namespaces.plone.org/supermodel/schema"
            xmlns:people="http://namespaces.plone.org/supermodel/people"
        >
            <schema>
                <people:columns>
                    <people:column>
                        <people:item>firstname</people:item>
                    </people:column>
                    <people:column>
                        <people:item>lastname</people:item>
                    </people:column>
                </people:columns>
                <field name="firstname" type="zope.schema.TextLine">
                    <title>First name</title>
                </field>
                <field name="lastname" type="zope.schema.TextLine">
                    <title>Last name</title>
                </field>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder(),
            )

            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model
            ).id

            api.content.create(
                id='miri', type=person, container=lst,
                firstname='Miriam', lastname='Linky'
            )
            api.content.create(
                id='zack', type=person, container=lst,
                firstname='Zack', lastname='Brown'
            )

            view = lst.unrestrictedTraverse('@@view')

        page = pq(view())

        self.assertEqual(len(page.find('thead th')), 2)

        self.assertEqual(len(page.find('tbody tr')), 2)
        self.assertEqual(len(page.find('tbody tr:first-child td')), 2)

        self.assertEqual(page.find('thead').text(), 'First name Last name')
        self.assertEqual(page.find('tbody').text(), 'Miriam Linky Zack Brown')

    def test_list_view_filter(self):

        model = """<?xml version='1.0' encoding='utf8'?>
        <model xmlns="http://namespaces.plone.org/supermodel/schema"
               xmlns:people="http://namespaces.plone.org/supermodel/people">
            <schema>
                <people:columns>
                    <people:column selectable="true">
                        <people:item>country</people:item>
                    </people:column>
                </people:columns>
                <field name="country" type="zope.schema.TextLine">
                    <title>Country</title>
                </field>
            </schema>
        </model>"""

        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder(),
            )

            country = self.new_temporary_type(
                behaviors=[IPerson.__identifier__],
                model_source=model
            ).id

            countries = ['Germany', 'Switzerland']

            for c in countries:
                api.content.create(
                    title=c, type=country, container=lst, country=c
                )

            view = lst.unrestrictedTraverse('@@view')

        columns = view.columns()

        self.assertEqual(len(view.people()), 2)
        self.assertEqual(len(columns), 1)
        self.assertEqual(view.selected_column_value(columns[0]), '__all__')
        self.assertEqual(view.column_values(columns[0]), countries)

        view.request['filter-country'] = 'Switzerland'

        self.assertEqual(len(view.people()), 1)
        self.assertEqual(view.column_values(columns[0]), countries)
        self.assertEqual(view.selected_column_value(columns[0]), 'Switzerland')

    def test_list_view_schema(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        # don't call if there are no people
        self.assertRaises(AssertionError, lambda: view.schema)

        with self.user('admin'):
            person = self.new_temporary_type(
                behaviors=[IPerson.__identifier__]
            )
            api.content.create(id='test', type=person.id, container=lst)

        self.assertEqual(view.schema, person.lookupSchema())

    def test_list_view_filter_property(self):
        with self.user('admin'):
            lst = api.content.create(
                id='test',
                type='seantis.people.list',
                container=self.new_temporary_folder()
            )

            view = lst.unrestrictedTraverse('@@view')

        self.assertEqual(view.filter, None)

        view.filter_prefix = 'test-filter-'
        view.request['test-filter-key'] = 'value'

        self.assertEqual(view.filter.key, 'key')
        self.assertEqual(view.filter.value, 'value')
        self.assertEqual(view.filter.title, u'')



from setuptools import setup, find_packages
import os

name = 'seantis.people'
description = (
    "A list of people, optionally organized by organisations and positions."
)
version = '0.6'

tests_require = [
    'plone.app.testing',
    'collective.betterbrowser[pyquery]',
    'seantis.plonetools[tests]',
    'mock'
]


def get_long_description():
    readme = open('README.rst').read()
    history = open(os.path.join('docs', 'HISTORY.rst')).read()
    contributors = open(os.path.join('docs', 'CONTRIBUTORS.rst')).read()

    # cut the part before the description to avoid repetition on pypi
    readme = readme[readme.index(description) + len(description):]

    return '\n'.join((readme, contributors, history))


setup(
    name=name, version=version, description=description,
    long_description=get_long_description(),
    classifiers=[
        'Framework :: Plone',
        'Framework :: Plone :: 4.3',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python',
    ],
    keywords='plone seantis people persons organizations positions',
    author='Seantis GmbH',
    author_email='info@seantis.ch',
    url='https://github.com/seantis/seantis.people',
    license='GPL v2',
    packages=find_packages(exclude=['ez_setup']),
    namespace_packages=['seantis'],
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'Plone>=4.3',
        'plone.api',
        'five.grok',
        'plone.behavior',
        'plone.app.dexterity [grok, relations]',
        'seantis.plonetools>=0.7',
        'collective.js.underscore',
        'collective.dexteritytextindexer',
        'tablib'
    ],
    extras_require=dict(
        tests=tests_require
    ),
    entry_points="""
    # -*- Entry points: -*-

    [z3c.autoinclude.plugin]
    target = plone
    """
)



from collections import defaultdict

from plone import api
from zope.interface import implements

from seantis.plonetools import tools
from seantis.people.interfaces import IPerson


class Person(object):
    implements(IPerson)

    def __init__(self, context):
        self.context = context

    def memberships(self):
        catalog = api.portal.get_tool(name='portal_catalog')
        memberships = catalog(portal_type='seantis.people.membership')

        return memberships

    def memberships_by_organization(self, organizations=None):
        result = defaultdict(list)

        for membership in self.memberships():
            organization = tools.get_parent(membership).title

            if organizations is None or organization in organizations:
                result[organization].append(membership)

        return result



from lxml import etree

from plone.supermodel.parser import ISchemaMetadataHandler

from plone.supermodel.utils import ns
from zope.interface import implements

from seantis.plonetools import tools

NAME_FROM_PERSON = u'seantis.people.name_from_person'
PERSON_COLUMNS = u'seantis.people.person_column'
PERSON_ORDER = u'seantis.people.order'
PERSON_SELECTABLE = u'seantis.people.selectable'
PERSON_DETAILS = u'seantis.people.details'

# Supermodel namespace and prefix
PEOPLE_NAMESPACE = 'http://namespaces.plone.org/supermodel/people'
PEOPLE_PREFIX = 'people'

missing = object()


def get_title_fields(schema):
    return schema.queryTaggedValue(NAME_FROM_PERSON, [])


def set_title_fields(schema, fields):
    schema.setTaggedValue(NAME_FROM_PERSON, sorted(
        list(set(fields)), key=fields.index  # unique values, preserving order
    ))


def get_selectable_fields(schema):
    return schema.queryTaggedValue(PERSON_SELECTABLE, [])


def set_selectable_fields(schema, fields):
    schema.setTaggedValue(PERSON_SELECTABLE, list(set(fields)))


def get_columns(schema):
    return schema.queryTaggedValue(PERSON_COLUMNS, [])


def set_columns(schema, columns):
    for column in columns:
        assert not isinstance(column, basestring), """
            A list of list is expected, not a list of strings.
        """
    schema.setTaggedValue(PERSON_COLUMNS, columns)

    for column in get_columns(schema):
        for field in column:
            tools.add_attribute_to_metadata(field)


def get_order(schema):
    return schema.queryTaggedValue(PERSON_ORDER, [])


def set_order(schema, order):
    schema.setTaggedValue(PERSON_ORDER, order)


def get_detail_fields(schema):
    return schema.queryTaggedValue(PERSON_DETAILS, {})


def set_detail_fields(schema, fields):
    schema.setTaggedValue(PERSON_DETAILS, fields)


class NodeHandler(object):

    namespace = PEOPLE_NAMESPACE
    prefix = PEOPLE_PREFIX

    def __init__(self, tagname):
        self.tagname = tagname
        etree.register_namespace(self.prefix, self.namespace)

    def prefixed(self, name):
        return ns(name, self.namespace)

    def nodes(self, node, name):
        return node.findall('./{}'.format(self.prefixed(name)))

    def tags(self, node):
        return self.nodes(node, self.tagname)


class ItemListHandler(NodeHandler):

    def __init__(self, tagname, getter, setter):
        super(ItemListHandler, self).__init__(tagname)
        self.getter = getter
        self.setter = setter

    def parse(self, schema_node, schema):
        tags = self.tags(schema_node)

        if not tags:
            return

        values = []

        for tag in tags:
            for item in self.nodes(tag, 'item'):
                values.append(item.text.strip())

        self.setter(schema, values)

    def write(self, schema_node, schema):
        values = self.getter(schema)

        if not values:
            return

        element = etree.Element(self.prefixed(self.tagname))
        for value in values:
            item = etree.Element(self.prefixed('item'))
            item.text = value

            element.append(item)

        schema_node.append(element)


class DetailsHandler(NodeHandler):

    def parse(self, schema_node, schema):
        tags = self.tags(schema_node)

        if not tags:
            return

        detail_fields = {}

        for tag in tags:
            position = tag.get('position') or 'left'
            detail_fields[position] = []

            for item in self.nodes(tag, 'item'):
                detail_fields[position].append(item.text.strip())

        set_detail_fields(schema, detail_fields)

    def write(self, schema_node, schema):
        detail_fields = get_detail_fields(schema)

        if not detail_fields:
            return

        for position, fields in detail_fields.items():
            element = etree.Element(self.prefixed(self.tagname))
            element.set('position', position)

            for field in fields:
                item = etree.Element(self.prefixed('item'))
                item.text = field

                element.append(item)

            schema_node.append(element)


class ColumnsHandler(NodeHandler):

    def parse(self, schema_node, schema):
        tags = self.tags(schema_node)

        if not tags:
            return

        columns = []
        selectable_fields = []

        for tag in tags:
            for column in self.nodes(tag, 'column'):
                items = self.nodes(column, 'item')

                if len(items) == 1 and column.get('selectable'):
                    selectable = True
                else:
                    selectable = False

                column = []
                for item in items:
                    value = item.text.strip()
                    column.append(value)

                    if selectable:
                        selectable_fields.append(value)

                columns.append(column)

        set_columns(schema, columns)
        set_selectable_fields(schema, selectable_fields)

    def write(self, schema_node, schema):
        columns = get_columns(schema)
        selectable_fields = get_selectable_fields(schema)

        if not columns:
            return

        columns_el = etree.Element(self.prefixed(self.tagname))

        for column in columns:
            column_el = etree.Element(self.prefixed('column'))

            if len(column) == 1 and column[0] in selectable_fields:
                column_el.set('selectable', 'true')

            for field in column:
                item_el = etree.Element(self.prefixed('item'))
                item_el.text = field
                column_el.append(item_el)

            columns_el.append(column_el)

        schema_node.append(columns_el)


class PeopleSchemaMetaHandler(object):

    implements(ISchemaMetadataHandler)

    namespace = PEOPLE_NAMESPACE
    prefix = PEOPLE_PREFIX

    handlers = [
        ItemListHandler('title', get_title_fields, set_title_fields),
        ItemListHandler('order', get_order, set_order),
        DetailsHandler('details'),
        ColumnsHandler('columns')
    ]

    def read(self, schema_node, schema):
        for handler in self.handlers:
            handler.parse(schema_node, schema)

    def write(self, schema_node, schema):
        for handler in self.handlers:
            handler.write(schema_node, schema)



import warnings

from collections import namedtuple

from five import grok
from plone import api
from plone.dexterity.content import Container
from Products.CMFPlone.interfaces.constrains import IConstrainTypes, ENABLED

from seantis.people.interfaces import IPerson, IList
from seantis.people.supermodel import get_selectable_field_ix
from seantis.plonetools import tools


ListFilter = namedtuple('ListFilter', ['key', 'value', 'title'])
LetterFilter = namedtuple('LetterFilter', ['value', 'title'])


class List(Container):

    def people(self, filter=None, batch_start=None, batch_size=None):
        catalog = api.portal.get_tool('portal_catalog')

        query = {}
        query['path'] = {
            'query': '/'.join(self.getPhysicalPath()), 'depth': 1
        }
        query['sort_on'] = 'sortable_title'
        query['sort_order'] = 'ascending'

        if batch_start is not None and batch_size is not None:
            query['b_start'] = batch_start
            query['b_size'] = batch_size

        if filter:
            used_type = self.used_type()

            if used_type:

                if isinstance(filter, ListFilter):
                    field_index = get_selectable_field_ix(
                        used_type.id, filter.key
                    )

                if isinstance(filter, LetterFilter):
                    field_index = 'first_letter'

                query[field_index] = filter.value

        return catalog(query)

    def letters(self):
        catalog = api.portal.get_tool('portal_catalog')
        index = catalog._catalog.getIndex('first_letter')

        letters = set()

        for person in self.people():
            letter = index.getEntryForObject(person.getRID())

            if letter:
                letters.add(letter)

        return sorted(letters, key=tools.unicode_collate_sortkey())

    def possible_types(self):
        return tools.get_type_info_by_behavior(IPerson.__identifier__)

    def available_types(self):
        used_type = self.used_type()

        if used_type:
            return [used_type]
        else:
            return self.possible_types()

    def used_type(self):
        catalog = api.portal.get_tool('portal_catalog')
        path = {'query': '/'.join(self.getPhysicalPath()), 'depth': 1}

        for fti in self.possible_types():
            if catalog(path=path, portal_type=fti.id, sort_limit=1)[:1]:
                return fti

        return None


class ListConstrainTypes(grok.Adapter):
    grok.provides(IConstrainTypes)
    grok.context(IList)

    def getConstrainTypesMode(self):
        return ENABLED

    def allowedContentTypes(self):
        return self.context.available_types()

    def getLocallyAllowedTypes(self):
        return [t.id for t in self.allowedContentTypes()]

    def getImmediatelyAddableTypes(self):
        return self.getLocallyAllowedTypes()
